from config import AIIAConfig
from torch import nn
import torch
import os
import copy  # Add this for deep copying

class AIIA(nn.Module):
    def __init__(self, config: AIIAConfig, **kwargs):
        super(AIIA, self).__init__()
        # Create a deep copy of the configuration to avoid sharing
        self.config = copy.deepcopy(config)
        
        # Update the config with any additional keyword arguments
        for key, value in kwargs.items():
            setattr(self.config, key, value)

    def save(self, path: str):
        if not os.path.exists(path):
            os.makedirs(path, exist_ok=True)
        torch.save(self.state_dict(), f"{path}/model.pth")
        self.config.save(path)

    @classmethod
    def load(cls, path):
        config = AIIAConfig.load(path)
        model = cls(config)
        model.load_state_dict(torch.load(f"{path}/model.pth"))
        return model

class AIIABaseShared(AIIA):
    """
    Base class with parameter sharing.
    All hidden layers share the same weights
    """
    def __init__(self, config: AIIAConfig, **kwargs):
        super().__init__(config=config, **kwargs)
        self.config = copy.deepcopy(config)
        
        # Update config with new parameters if provided
        for key, value in kwargs.items():
            setattr(self.config, key, value)

        # Initialize shared layers
        self.conv_layer = nn.Conv2d(
            self.config.num_channels,
            self.config.hidden_size,
            kernel_size=self.config.kernel_size,
            padding=1
        )
        self.activation_function = getattr(nn, self.config.activation_function)()
        self.max_pool = nn.MaxPool2d(kernel_size=2)
        
        # Create a Sequential container with shared layers repeated
        layers = []
        for _ in range(self.config.num_hidden_layers):
            layers.extend([
                self.conv_layer,
                self.activation_function,
                self.max_pool
            ])
            
        self.cnn = nn.Sequential(*layers)

    def forward(self, x):
        return self.cnn(x)
    

class AIIABase(AIIA):
    def __init__(self, config: AIIAConfig, **kwargs):
        super().__init__(config=config, **kwargs)  
        self.config = self.config
        
        # Initialize layers based on configuration
        layers = []
        in_channels = self.config.num_channels
        
        for _ in range(self.config.num_hidden_layers):
            layers.extend([
                nn.Conv2d(in_channels, self.config.hidden_size,
                          kernel_size=self.config.kernel_size, padding=1),
                getattr(nn, self.config.activation_function)(),
                nn.MaxPool2d(kernel_size=2)
             ])
            in_channels = self.config.hidden_size
        
        self.cnn = nn.Sequential(*layers)

    def forward(self, x):
        return self.cnn(x)

class AIIAExpert(AIIA):
    def __init__(self, config: AIIAConfig, base_class=AIIABase, **kwargs):
        super().__init__(config=config, **kwargs)
        self.config = self.config
        
        # Initialize base CNN with configuration and chosen base class
        if issubclass(base_class, AIIABase):
            self.base_cnn = AIIABase(self.config, **kwargs)
        elif issubclass(base_class, AIIABaseShared):
            self.base_cnn = AIIABaseShared(self.config, **kwargs)
        else:
            raise ValueError("Invalid base class")

class AIIAmoe(AIIA):
    def __init__(self, config: AIIAConfig, num_experts: int = 3, base_class=AIIABase, **kwargs):
        super().__init__(config=config, **kwargs)
        self.config = self.config
        
        # Update config with new parameters if provided
        self.config.num_experts = num_experts
        
        # Initialize multiple experts using chosen base class
        self.experts = nn.ModuleList([
            AIIAExpert(self.config, base_class=base_class, **kwargs) 
            for _ in range(self.config.num_experts)
        ])
        
        # Create gating network
        self.gate = nn.Sequential(
            nn.Linear(self.config.hidden_size, self.config.num_experts),
            nn.Softmax(dim=1)
        )

    def forward(self, x):
        expert_outputs = torch.stack([expert(x) for expert in self.experts], dim=1)
        gate_weights = self.gate(torch.mean(expert_outputs, (2, 3)))
        merged_output = torch.sum(
            expert_outputs * gate_weights.unsqueeze(2).unsqueeze(3), dim=1
        )
        return merged_output

class AIIAchunked(AIIA):
    def __init__(self, config: AIIAConfig, patch_size: int = 16, base_class=AIIABase, **kwargs):
        super().__init__(config=config, **kwargs)  
        self.config = self.config
        
        # Update config with new parameters if provided
        self.config.patch_size = patch_size
        
        # Initialize base CNN for processing each patch using the specified base class
        if issubclass(base_class, AIIABase):
            self.base_cnn = AIIABase(self.config, **kwargs)
        elif issubclass(base_class, AIIABaseShared):  # Add support for AIIABaseShared
            self.base_cnn = AIIABaseShared(self.config, **kwargs)
        else:
            raise ValueError("Invalid base class")

    def forward(self, x):
        patches = x.unfold(2, self.patch_size, self.patch_size).unfold(3, self.patch_size, self.patch_size)
        patches = patches.contiguous().view(patches.size(0), patches.size(1), -1, self.patch_size, self.patch_size)
        patch_outputs = []
        
        for p in torch.split(patches, 1, dim=2):
            p = p.squeeze(2)
            po = self.base_cnn(p)
            patch_outputs.append(po)
            
        combined_output = torch.mean(torch.stack(patch_outputs, dim=0), dim=0)
        return combined_output

class AIIArecursive(AIIA):
    def __init__(self, config: AIIAConfig, recursion_depth: int = 3, base_class=AIIABase, **kwargs):
        
        super().__init__(config=config, **kwargs)
        self.config = self.config
        
        # Pass recursion_depth as a kwarg to the config
        self.config.recursion_depth = recursion_depth
            
        # Initialize chunked CNN with updated config
        self.chunked_cnn = AIIAchunked(self.config, base_class, **kwargs)
        
    def forward(self, x, depth=0):
        if depth == self.recursion_depth:
            return self.chunked_cnn(x)
        else:
            patches = x.unfold(2, 16, 16).unfold(3, 16, 16)
            patches = patches.contiguous().view(patches.size(0), patches.size(1), -1, 16, 16)
            processed_patches = []
            
            for p in torch.split(patches, 1, dim=2):
                p = p.squeeze(2)
                pp = self.forward(p, depth + 1)
                processed_patches.append(pp)
                
            combined_output = torch.mean(torch.stack(processed_patches, dim=0), dim=0)
            return combined_output